/*
    PPGNUstepGlue_WindowStyleOffsets.m

    Copyright 2014-2016 Josh Freeman
    http://www.twilightedge.com

    This file is part of PikoPixel for GNUstep.
    PikoPixel is a graphical application for drawing & editing pixel-art images.

    PikoPixel is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License version 3,
    as published by the Free Software Foundation.

    PikoPixel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// - Workarounds for window style offsets issues on Compiz WM (titlebars obscure content)
// & Openbox WM (drawing artifacts after maximizing or switching resizeable/nonresizable)
//
// - Force GNUstep to ignore any cached style offset values stored in the root window, instead
// check offsets manually (in case the window manager changed since the values were cached)

#ifdef GNUSTEP

#import <Cocoa/Cocoa.h>
#import "NSObject_PPUtilities.h"
#import "PPAppBootUtilities.h"
#import "PPGNUstepGlueUtilities.h"


#define kUserDefaultsKey_FallbackWindowStyleOffsets     @"GSGlue_FallbackWindowStyleOffsets"


typedef unsigned long int PPXWindow; // local definition of libX11's Window type


@interface NSUserDefaults (PPGNUstepGlue_WindowStyleOffsetsUtilities)

+ (void) ppGSGlue_LoadFallbackWindowStyleOffsets_left: (float *) left
            right: (float *) right
            top: (float *) top
            bottom: (float *) bottom;

+ (void) ppGSGlue_SaveFallbackWindowStyleOffsets_left: (float) left
            right: (float) right
            top: (float) top
            bottom: (float) bottom;

@end


@implementation NSObject (PPGNUstepGlue_WindowStyleOffsets)

+ (void) ppGSGlue_WindowStyleOffsets_InstallPatches_Compiz
{
    [NSClassFromString(@"XGServer")
        ppSwizzleInstanceMethodWithSelector: @selector(styleoffsets::::::)
        forNSObjectInstanceMethodWithSelector: @selector(ppGSPatch_Compiz_Styleoffsets::::::)];
}

+ (void) ppGSGlue_WindowStyleOffsets_InstallPatches_Openbox
{
    [NSClassFromString(@"XGServer")
        ppSwizzleInstanceMethodWithSelector: @selector(styleoffsets::::::)
        forNSObjectInstanceMethodWithSelector: @selector(ppGSPatch_Openbox_Styleoffsets::::::)];
}

+ (void) ppGSGlue_WindowStyleOffsets_Install
{
    if (PPGSGlueUtils_WindowManagerMatchesTypeMask(kPPGSWindowManagerTypeMask_Compiz))
    {
        [self ppGSGlue_WindowStyleOffsets_InstallPatches_Compiz];
    }
    else if (PPGSGlueUtils_WindowManagerMatchesTypeMask(kPPGSWindowManagerTypeMask_Openbox))
    {
        [self ppGSGlue_WindowStyleOffsets_InstallPatches_Openbox];
    }
}

+ (void) load
{
    macroPerformNSObjectSelectorAfterAppLoads(ppGSGlue_WindowStyleOffsets_Install);

    PPGSGlueUtils_PerformNSUserDefaultsSelectorBeforeGSBackendLoads(
                                            @selector(ppGSGlue_IgnoreRootWindowStyleOffsets));
}

// PATCH: -[XGServer styleoffsets::::::] (Compiz)
// When running on Compiz window manager, styleoffsets:::::: method will return garbage values
// for random styles (and which style values return garbage will change each time the app runs);
// Workaround patch tries to establish a set of valid offset defaults by walking through all
// nonzero style values (1-15) until it finds a set that seems reasonable (nonzero value for t);
// If the method's returned values appear to be garbage, it replaces them with the previously
// found valid default offsets

- (void) ppGSPatch_Compiz_Styleoffsets: (float *) l : (float *) r : (float *) t : (float *) b
		     : (unsigned int) style : (PPXWindow) win
{
    [self ppGSPatch_Compiz_Styleoffsets: l : r : t : b : style : win];

    if (style & NSTitledWindowMask)
    {
        // initial fallback values {0,0,28,0} are defaults for titled Compiz windows on Ubuntu
        static float fallbackOffset_Left = 0, fallbackOffset_Right = 0, 
                        fallbackOffset_Top = 28, fallbackOffset_Bottom = 0;
        static bool didLoadSavedFallbackOffsets = NO, didUpdateSavedFallbackOffsets = NO;
        bool styleOffsetsAreInvalid;

        styleOffsetsAreInvalid = (*t <= 0) ? YES : NO;

        if (styleOffsetsAreInvalid)
        {
            if (!didLoadSavedFallbackOffsets)
            {
                [NSUserDefaults 
                        ppGSGlue_LoadFallbackWindowStyleOffsets_left: &fallbackOffset_Left
                                                                right: &fallbackOffset_Right
                                                                top: &fallbackOffset_Top
                                                                bottom: &fallbackOffset_Bottom];

                didLoadSavedFallbackOffsets = YES;
            }

            *l = fallbackOffset_Left;
            *r = fallbackOffset_Right;
            *t = fallbackOffset_Top;
            *b = fallbackOffset_Bottom;
        }
        else if (!didUpdateSavedFallbackOffsets)
        {
            if ((*l != fallbackOffset_Left)
                || (*r != fallbackOffset_Right)
                || (*t != fallbackOffset_Top)
                || (*b != fallbackOffset_Bottom))
            {
                fallbackOffset_Left = *l;
                fallbackOffset_Right = *r;
                fallbackOffset_Top = *t;
                fallbackOffset_Bottom = *b;

                [NSUserDefaults 
                        ppGSGlue_SaveFallbackWindowStyleOffsets_left: fallbackOffset_Left
                                                                right: fallbackOffset_Right
                                                                top: fallbackOffset_Top
                                                                bottom: fallbackOffset_Bottom];
            }          

            didUpdateSavedFallbackOffsets = YES;
        }
    }
}

// PATCH: -[XGServer styleoffsets::::::] (Openbox)
// When running on Openbox window manager, maximizing a titled window or switching a titled
// window between resizable & non-resizable causes drawing artifacts due to incorrect style
// offsets - this is because Openbox window decorations are different sizes for different
// window states, and when the decoration sizes change on-the-fly, they no longer line up with
// GNUstep's drawing/graphics state (which seems to use cached offset values);
// Patch sets the win parameter to zero (if the window style is titled) before calling the
// original styleoffsets:::::: implementation - this forces it to return cached offset values
// (which should match the window's initial state & GNUstep's) instead of querying the window
// directly for its current offsets (which may no longer match GNUstep's state)

- (void) ppGSPatch_Openbox_Styleoffsets: (float *) l : (float *) r : (float *) t : (float *) b
		     : (unsigned int) style : (PPXWindow) win
{
    if (style & NSTitledWindowMask)
    {
        win = 0;
    }

    [self ppGSPatch_Openbox_Styleoffsets: l : r : t : b : style : win];
}

@end

@implementation NSUserDefaults (PPGNUstepGlue_WindowStyleOffsets)

- (void) ppGSGlue_IgnoreRootWindowStyleOffsets
{
    NSDictionary *defaultsDict = 
                            [NSDictionary dictionaryWithObject: [NSNumber numberWithBool: YES]
                                            forKey: @"GSIgnoreRootOffsets"];

    if (!defaultsDict)
        goto ERROR;

    [self registerDefaults: defaultsDict];

    return;

ERROR:
    return;
}

+ (void) ppGSGlue_LoadFallbackWindowStyleOffsets_left: (float *) left
            right: (float *) right
            top: (float *) top
            bottom: (float *) bottom
{
    NSUserDefaults *userDefaults;
    NSArray *fallbackOffsets;

    if (!left || !right || !top || !bottom)
    {
        goto ERROR;
    }

    userDefaults = [NSUserDefaults standardUserDefaults];

    fallbackOffsets = [NSArray arrayWithObjects: 
                                    [NSNumber numberWithFloat: *left],
                                    [NSNumber numberWithFloat: *right],
                                    [NSNumber numberWithFloat: *top],
                                    [NSNumber numberWithFloat: *bottom],
                                    nil];

    if (fallbackOffsets)
    {
        [userDefaults registerDefaults: 
                        [NSDictionary dictionaryWithObject: fallbackOffsets
                                        forKey: kUserDefaultsKey_FallbackWindowStyleOffsets]];
    }
    
    fallbackOffsets = [userDefaults objectForKey: kUserDefaultsKey_FallbackWindowStyleOffsets];

    if (![fallbackOffsets isKindOfClass: [NSArray class]]
        || ([fallbackOffsets count] < 4))
    {
        goto ERROR;
    }

    *left = [[fallbackOffsets objectAtIndex: 0] floatValue];
    *right = [[fallbackOffsets objectAtIndex: 1] floatValue];
    *top = [[fallbackOffsets objectAtIndex: 2] floatValue];
    *bottom = [[fallbackOffsets objectAtIndex: 3] floatValue];

    return;

ERROR:
    return;
}

+ (void) ppGSGlue_SaveFallbackWindowStyleOffsets_left: (float) left
            right: (float) right
            top: (float) top
            bottom: (float) bottom
{
    NSUserDefaults *userDefaults = [NSUserDefaults standardUserDefaults];
    NSArray *fallbackOffsets = [NSArray arrayWithObjects: 
                                            [NSNumber numberWithFloat: left],
                                            [NSNumber numberWithFloat: right],
                                            [NSNumber numberWithFloat: top],
                                            [NSNumber numberWithFloat: bottom],
                                            nil];

    if (!fallbackOffsets)
        goto ERROR;

    [userDefaults setObject: fallbackOffsets
                    forKey: kUserDefaultsKey_FallbackWindowStyleOffsets];

    return;

ERROR:
    return;
}

@end

#endif  // GNUSTEP

